#include "VoxGrid.h"

#include "Mesh.h"

#include "../voxelizer/meshVoxelizer.h" // mesh voxelizer (Sylvain Lefebvre's courtesy)
#include "../qpsolver/MOSEK_solver.h" // QP solver (Mosek library)

	VoxGrid::VoxGrid(const int res)
	{
		m_res = res;
		m_lowerLeft.setZero();
		m_upperRight.setConstant(1.0f);
	
		m_boxArray.init(res, res, res);
		m_nodeArray.init(res+1, res+1, res+1);

		m_frac = (m_upperRight-m_lowerLeft).cwiseQuotient(RowVector3(res,res,res));
	}

	void VoxGrid::initVoxels(const Mesh & mesh)
	{
		m_boxArray.setAllTo(-1);
		nnzBoxes = 0;
		meshVoxelize(mesh,m_res,*this); // call voxelizer
	
		// verify if all the voxels containing a mesh vertices are included (usually voxelizer do not guarantee this property)
		for(int i = 0 ; i < mesh.nbV ; ++i) {
			RowVector3i t = RowVector3i::Zero();
			if(getBoxCoordsContainingPoint(mesh.transformedPosition(i),t)) {
				if(m_boxArray(t[0],t[1],t[2]) == -1) {
					addVoxel(t[0],t[1],t[2]);
				}
			}
			else {
				cout << "Error: the mesh vertex is outside of the voxel grid" << endl;
			}
		}

	}

	void VoxGrid::addVoxel(int x, int y, int z) 
	{
		m_boxArray(x,y,z) = nnzBoxes++;
	}

	void VoxGrid::initStructure()
	{
		// a node is included if at least one of the surrounding box is an occupied voxel
		m_nodeArray.setAllTo(-1);
		nnzNodes = 0;
		for (int x=0; x<m_res+1; x++) {
			for (int y=0; y<m_res+1; y++) {
				for (int z=0; z<m_res+1; z++) {
					bool occupiedNeighboringBox = false;
					for (int dx=-1; dx<1; dx++) {
						for (int dy=-1; dy<1; dy++) {
							for (int dz=-1; dz<1; dz++) {
								if (m_boxArray.validIndices(x + dx, y + dy, z + dz)) {
									if (m_boxArray(x + dx, y + dy, z + dz) != -1) occupiedNeighboringBox = true;								
								}
							}
						}
					}				

					if (occupiedNeighboringBox) m_nodeArray(x, y, z) = nnzNodes++;
				}
			}
		}
		
		// set node-node neighboring indexes
		m_nodeNodes.setConstant(nnzNodes,6,-1);
		for (int x=0; x<m_res+1; x++) {
			for (int y=0; y<m_res+1; y++) {
				for (int z=0; z<m_res+1; z++) {
					if(m_nodeArray(x,y,z) == -1) continue;

					int idNode = m_nodeArray(x,y,z);

					int k = 0;
					for (int dw=-1; dw<=1; dw+=2) {
						if (m_nodeArray.validIndices(x + dw, y, z)) {
							m_nodeNodes(idNode,k) = m_nodeArray(x + dw, y, z);								
						}
						k++;
						if (m_nodeArray.validIndices(x, y + dw, z)) {
							m_nodeNodes(idNode,k) = m_nodeArray(x, y + dw, z);								
						}
						k++;
						if (m_nodeArray.validIndices(x, y, z + dw)) {
							m_nodeNodes(idNode,k) = m_nodeArray(x, y, z + dw);								
						}
						k++;
					}
				}
			}
		}

		// set box-node neighboring indexes
		m_boxNodes.setConstant(nnzBoxes, 8, -1);
		for (int x=0; x<m_res; x++) {
			for (int y=0; y<m_res; y++) {
				for (int z=0; z<m_res; z++) {
					const int idBox = m_boxArray(x,y,z);
					if (idBox == -1) continue;
				
					int cnt = 0;
					for (int dx=0; dx<2; dx++) {
						for (int dy=0; dy<2; dy++) {
							for (int dz=0; dz<2; dz++) {
								m_boxNodes(idBox,cnt++) = m_nodeArray(x + dx, y + dy, z + dz);
							}
						}
					}
				}
			}
		}

	}

	int VoxGrid::getNodeId(const RowVector3i & id) const
	{
		return m_nodeArray(id[0],id[1],id[2]);
	}

	RowVector3 VoxGrid::getNodePose(const RowVector3i & id) const
	{
		return m_lowerLeft + m_frac.cwiseProduct(RowVector3(id[0],id[1],id[2]));
	}

	int VoxGrid::getBoxContainingPoint(const RowVector3 & P, RowVector3 & t) const
	{
		if(P.cwiseMin(m_lowerLeft) != m_lowerLeft || P.cwiseMax(m_upperRight) != m_upperRight) {
			cout << "Error : point is outside the voxelgrid" << endl;
			return -1;
		}
		RowVector3 floatIndices = (P - m_lowerLeft).cwiseQuotient(m_frac);
		RowVector3i intIndices((int)floor(floatIndices[0]),(int)floor(floatIndices[1]),(int)floor(floatIndices[2]));
	
		if(!m_boxArray.validIndices(intIndices[0],intIndices[1],intIndices[2])) {
			cout << "Error : position in voxelgrid incorrect" << endl;
			return -1;
		}
		t = floatIndices - RowVector3(intIndices[0],intIndices[1],intIndices[2]);
	
		return m_boxArray(intIndices[0],intIndices[1],intIndices[2]);
	}

	bool VoxGrid::getBoxCoordsContainingPoint(const RowVector3 & P, RowVector3i & t) const
	{
		if(P.cwiseMin(m_lowerLeft) != m_lowerLeft || P.cwiseMax(m_upperRight) != m_upperRight) {
			cout << "Error : point is outside the voxelgrid" << endl;
			return false;
		}
		RowVector3 floatIndices = (P - m_lowerLeft).cwiseQuotient(m_frac);
		t = RowVector3i((int)floor(floatIndices[0]),(int)floor(floatIndices[1]),(int)floor(floatIndices[2]));
	
		if(!m_boxArray.validIndices(t[0],t[1],t[2])) {
			cout << "Error : position in voxelgrid incorrect" << endl;
			return false;
		}
		return true;
	}

	RowVector3i VoxGrid::getNodeClosestToPoint(const RowVector3 & P) const
	{

		if(P.cwiseMin(m_lowerLeft) != m_lowerLeft || P.cwiseMax(m_upperRight) != m_upperRight) {
			cout << "Error : point is outside the voxelgrid" << endl;
		}
		RowVector3 floatIndices = (P - m_lowerLeft).cwiseQuotient(m_frac);
		RowVector3i intIndices((int)floor(floatIndices[0]),(int)floor(floatIndices[1]),(int)floor(floatIndices[2]));
	
		if(!m_boxArray.validIndices(intIndices[0],intIndices[1],intIndices[2])) {
			cout << "Error : position in voxelgrid incorrect" << endl;
		}
		RowVector3 t = floatIndices - RowVector3(intIndices[0],intIndices[1],intIndices[2]);
		int dx = (t[0]<=0.5)?0:1;
		int dy = (t[1]<=0.5)?0:1;
		int dz = (t[2]<=0.5)?0:1;
	
		return RowVector3i(intIndices[0]+dx,intIndices[1]+dy,intIndices[2]+dz);
	}

	void VoxGrid::laplacianMEL(vector<SparseMatrixTriplet> &MEL) const
	{	
		MEL.clear();	
		vector<int> valences;
		valences.assign(getNumNodes(), 0);
		for(int v1 = 0 ; v1<getNumNodes() ; ++v1) {
			for(int k = 0 ; k < 6 ; ++k) {
				int v2 = m_nodeNodes(v1,k);
				if(v2 != -1) {
					MEL.push_back(SparseMatrixTriplet(v1,v2,-1.0f));
					valences[v1]++;
				}
			}
		}
		for (int v=0; v<getNumNodes(); v++) {
			MEL.push_back(SparseMatrixTriplet(v, v, valences[v]));
		}
	}

	void VoxGrid::getInterpolatedBBW(const RowVector3 & P, Weights & deformInfo, const int nbWeights) const
	{
		RowVector3 t;
		int idBox = getBoxContainingPoint(P,t);
		if(idBox == -1) {
			cout << "Error cannot create BBW for this mesh vertex" << endl;
			return;
		}

		for(int j = 0 ; j < nbWeights ; ++j) {
			float wj = 0.0f;
			for(int i = 0 ; i < 8 ; ++i) {
				const int idNode = m_boxNodes(idBox,i);
				float alpha = 1.0f;
				alpha *= (((i/4) % 2 == 0)?(1.0f-t[0]):t[0]);
				alpha *= (((i/2) % 2 == 0)?(1.0f-t[1]):t[1]);
				alpha *= ((i % 2 == 0)?(1.0f-t[2]):t[2]);
				wj += alpha * m_weights[idNode].getCoord(j);
			}
			deformInfo.pushWeight(wj);
		}
	}

	void VoxGrid::computeBBW(const Handles & handles)
	{	
		int N = getNumNodes();
		int M = handles.getNumHandles();
		
		// each voxel node will have weights
		m_weights.resize(N);

		// compute the Laplacian matrix
		SparseMatrix L, L2;
		vector<SparseMatrixTriplet> L_MEL;
		laplacianMEL(L_MEL);
		L.resize(N, N);
		L.setFromTriplets(L_MEL.begin(),L_MEL.end());	
		L2 = L*L;

		// compute the constraint matrix (each row corresponds to one handle)
		SparseMatrix A(M,N);
		vector<SparseMatrixTriplet> A_MEL;
		A_MEL.reserve(M);
		for(int i = 0 ; i < M ; ++i) {
			A_MEL.push_back(SparseMatrixTriplet(i, handles.getConstraint(i), 1.0));
		}
		A.setFromTriplets(A_MEL.begin(), A_MEL.end());
	
		VectorX x;
		VectorX b;
		MOSEKinterface mi0;
		MatrixXX zeros0(L2.rows(), 1); zeros0.setZero();

		for(int j = 0 ; j < M ; ++j) { // for each handle we compute the weight
			b.setZero(M,1);
			b[j] = 1.0f;

			x.setZero(N,1);
			mi0.solveQP_BBW_type(x, L2, zeros0, A, b, 1, MOSEKinterface::PRINT_LOG); // call Mosek QP solver

			for(int i = 0 ; i < N ; i++) { // we push the j-th weight for each node
				m_weights[i].pushWeight((ScalarType) x[i]);
			}
		}

		// for each node we now have all the weight so we can normalize them
		for(int i = 0 ; i < N; i++) {
			m_weights[i].normalizeWeights();
		}

	}
	
