#ifndef __VOXGRID_H
#define __VOXGRID_H

#include "Array3D.h"
#include "Weights.h"
#include "Handles.h"

class Mesh;

// class to store the voxel grid information
class VoxGrid {

public:
	VoxGrid(const int res); // initialize the voxel grid size and data structure
	~VoxGrid() {
		m_nodeArray.free();
		m_boxArray.free();
	}
	
	void addVoxel(int x, int y, int z); // add a voxel
	void initVoxels(const Mesh & mesh); // voxelize a mesh
	void initStructure(); // initialize neighboring data structures

	int getNumNodes() const { return nnzNodes; } // nb voxels
	int getNodeId(const RowVector3i & id) const;
	RowVector3 getNodePose(const RowVector3i & id) const;
	
	void computeBBW(const Handles & handles); // compute BBW using QP solver

	void getInterpolatedBBW(const RowVector3 & P, Weights & w, const int nbWeights) const; // return trilinearly interpolated weights for a 3D point inside the voxel grid
	
	RowVector3i getNodeClosestToPoint(const RowVector3 & P) const;

	void laplacianMEL(vector<SparseMatrixTriplet> &MEL) const; // compute elements of Laplacian matrix of the voxel grid (sparse matrix)

protected:
	int m_res; // number of boxes in x,y,z dimensions
	RowVector3 m_lowerLeft, m_upperRight; // placement in 3D space
	RowVector3 m_frac; // diagonal vector of a voxel
	
	int nnzBoxes, nnzNodes; // nb voxel boxes and nodes

	// all 3 following 3D arrays store either -1 if empty or element index (vertex, box)
	Array3Di m_nodeArray; // nodes 
	Array3Di m_boxArray; // boxes
	
	MatrixX8i m_boxNodes; // nnzBoxes x 8 int matrix of node indices (incident to a given box)
	MatrixX6i m_nodeNodes; // nnzNodes x 6 int matrix of node indices (incident to a given node)
		
	vector<Weights> m_weights; // list of the weights for each voxel node

	int getBoxContainingPoint(const RowVector3 & P, RowVector3 & t) const;
	bool getBoxCoordsContainingPoint(const RowVector3 & P, RowVector3i & t) const;

};

#endif
